//- =======================================================================
//+ GPU Scan based on the BPLG library LF
//- =======================================================================

//---- Header Dependencies -----------------------------------------------
//#include "BPLG_Scan.hxx"
#include "tools/cudacl.hxx"
#include "inc/op_copy.hxx"
#include "inc/op_twiddle.hxx"


#include "inc/op_shfl.hxx"

//---- Include Section ---------------------------------------------------
#include <cstdio>

#ifndef __CUDA_ARCH__
    #define __CUDA_ARCH__ CUDA_ARCH
#endif


#if  __CUDA_ARCH__ < 400
	#define tabla7 triXf32A
#endif

#if  __CUDA_ARCH__ >= 400
	#define tabla7 triXf32B
#endif


//---- Butterfly operator ------------------------------------------------

template<class DTYPE, int size> struct butterfly {

	//- Generic butterfly step, inclusive scan
	static inline __device__ void
	inc(DTYPE* data) {
		#pragma unroll
		for(int i = 1; i < size; i++)
			data[i] += data[i-1];
	}

	//- Generic butterfly step, exclusive scan
	static inline __device__ void
	exc(DTYPE* data){
		DTYPE acc = 0;
		#pragma unroll
		for(int i = 0; i < size; i++) {
			DTYPE tmp = data[i];
			data[i] = acc;
			acc += tmp;
		}
	}

};


//---- Radix operator ----------------------------------------------------

//- Normal radix stage, performs a normal inclusive scan
template<int RAD, class DTYPE> inline __device__ void
radix(DTYPE* data) {
	butterfly<DTYPE, RAD>::inc(data);
}

//- Mixed-radix stage, only called once before the main loop
template<int SIZE, int RAD, class DTYPE> inline __device__ void
radix(DTYPE* data) {
	#pragma unroll
	for(int i = 0; i < SIZE; i+= RAD)
		butterfly<DTYPE, RAD>::inc(data + i);
}

//- Generic radix stage, used in the main loop of the algorithm
template<int RAD, class DTYPE> inline __device__ void
radix(DTYPE* data, const DTYPE* rep, int stride, const DTYPE& base = 0) {
	DTYPE acc = base;
	#pragma unroll
	for(int i = 0; i < RAD; i++) {
		data[i] += acc;
		acc += rep[stride * i];
	}
}

//------- Cuda Kernels ---------------------------------------------------


template<int N, int DIR, int RAD, int SHM> __global__ void
KScanLF2(const float4* __restrict__ src, float4* dst, int stride)
{
	// Obtain group-1D, thread-X and batch-Y identifiers
	int groupId = get_group_id(0) + get_group_id(1) * get_num_groups(0);
	int threadId = get_local_id(0);      // Thread horizontal (N)
	int batchId = get_local_id(1);       // Thread vertical (batch)
	int thread1D = threadId + batchId * N/4;

	const int warp_size = 32;		
	// Offset for accessing thread data
	int shmOffset = batchId * warp_size;            // ShMem batch offset
       // Read Stride
	int glbRPos = (groupId * SHM)/4 + thread1D; // Read Pos
	
	// Statically allocate registers and shared memory
	float4 reg;
	
	__shared__ float shm[warp_size*(SHM/N)]; 
	reg = src[glbRPos];
	
	reg.y+=reg.x;
	reg.z+=reg.y;
	reg.w+=reg.z;
	

	
	
	int warp_id = threadIdx.x / warp_size;



	int i = 1;
	float n1 = 0;

	#pragma unroll
	for(int width = 2; width<=warp_size;width*=2)
	{
		int lane_id = (threadId &(width-1));
		float n = shfl<float>::shflDTYPE(reg.w, i-1, width);
		if(lane_id > (i-1)){
			reg.w+=n;
			n1+=n;		
		}	
		i*=2;

	}

	reg.x+=n1; reg.y+= n1; reg.z+=n1;

	

	if( (threadId & (warp_size-1))== (warp_size -1) )
		shm[warp_id+shmOffset]=reg.w;
	
	__syncthreads();
	
	
	if(!warp_id)
	{
		float warp_sum = 0;
		int lane_id =1 ;
		
		warp_sum = shm[shmOffset+threadId];
		float initial_sum = warp_sum;
		i=1;
		#pragma unroll 
		for(int width=2; width<=warp_size;width*=2)
		{
			lane_id = (threadId & (width-1));
			float n = shfl<float>::shflDTYPE(warp_sum, i-1, width);
			if (lane_id> (i-1))
				warp_sum+= n;
			i*=2;
		}	
		
		shm[shmOffset+threadId]= warp_sum-initial_sum;
	}			
	__syncthreads();

	reg.x+=shm[shmOffset+warp_id];
	reg.y+=shm[shmOffset+warp_id];
	reg.z+=shm[shmOffset+warp_id];
	reg.w+=shm[shmOffset+warp_id];



	// Store the final result in global memory
		
	dst[glbRPos]=reg;
}









// --- BranchTable -------------------------------------------------------



//- Declare kernel prototype for BranchTables
struct kernelCfg { 
	void(*kerPtr)(const float4*, float4*, int);
	short R;                         // Radix 
	short S;                         // Shared Memory
};

//- Initializes a null configuration for unsupported sizes
#define NULL_ROW(_N) { NULL, 0, 0 }

//- Initializes a configuration entry with the specified parameters
#define ROW(_Kernel, _N, _S, _R) { \
	_Kernel< _N, 1, _R, (_N > _S ? _N : _S)>, \
	_R, (_N > _S ? _N : _S) }



//- Template instantiation and branchtable for 'float' kernels
const static kernelCfg triXf32A[] = { //! GPU dependent
	NULL_ROW(1),
	NULL_ROW(2),
	ROW(KScanLF2,    4, 128,  4),
	ROW(KScanLF2,    8, 128,  4),
	ROW(KScanLF2,   16, 256,  4),
	ROW(KScanLF2,   32, 512,  4),
	ROW(KScanLF2,   64, 2048,  4),
	ROW(KScanLF2,  128, 1024,  4),
	ROW(KScanLF2,  256, 1024,  4),
	ROW(KScanLF2,  512, 1024,  4),
	ROW(KScanLF2, 1024,1024,  4),
	ROW(KScanLF2, 2048,2048,  4),
	ROW(KScanLF2, 4096,4096,  4),
	NULL_ROW(16384),
};

const static kernelCfg triXf32B[] = { //! GPU dependent
	NULL_ROW(1),
	NULL_ROW(2),
	ROW(KScanLF2,    4, 128,  4),
	ROW(KScanLF2,    8, 128,  4),
	ROW(KScanLF2,   16, 256,  4),
	ROW(KScanLF2,   32, 512,  4),
	ROW(KScanLF2,   64, 2048,  4),
	ROW(KScanLF2,  128, 1024,  4),
	ROW(KScanLF2,  256, 1024,  4),
	ROW(KScanLF2,  512, 1024,  4),
	ROW(KScanLF2, 1024,1024,  4),
	ROW(KScanLF2, 2048,2048,  4),
	ROW(KScanLF2, 4096,4096,  4),
	NULL_ROW(16384),
};


// --- Launcher ----------------------------------------------------------

//- Kernel launcher: Configures threads and blocks before launch
template<typename T> int
KLauncher(const kernelCfg* cfgTable, int tableBytes,
	T* srcPtr, T* dstPtr, int dir, int N, int batch) {

	

	// Check configuration and data pointers are valid
	if(!cfgTable || !srcPtr || !dstPtr || !batch)
		return printf("* FIXME: Bad launcher parameters *\n"), -1;

	



	// Verify that the selected N is in the table range
	const int TableSize = tableBytes / sizeof(kernelCfg);
	const int log2N = Log2(N);
	if(log2N < 0 || log2N >= TableSize) return -1;	

	

	// Obtains the corresponding kernel configuration
	const kernelCfg launchCfg = cfgTable[log2N];
	if(!launchCfg.kerPtr) return -1;

	


	// Check for valid input size
	const int vSize = N * batch;   //! For tridiagonal, dim M=4
	if((launchCfg.S - 1) & vSize)  // previously: launchCfg.S > vSize
		return printf("* FIXME: Input size not divisible by %i *\n", launchCfg.S), -1;

	// Configure thread geometry for kernel launch
	int threadsX, threadsY, threadsZ = 1;
	threadsX = N / launchCfg.R;    // Threads colaborando
	threadsY = launchCfg.S / N;    // Threads en batch
	dim3 threadsPerBlock(threadsX, threadsY, threadsZ);

	// Valite thread geometry configuration
	const int MaxThreads = 1024;   // Max threads per block
	int numThreads = threadsX * threadsY * threadsZ;
	
	if(numThreads < 1 || numThreads > MaxThreads || threadsZ > 64) return -1;

	

	// Obtain required total blocks
	int numBlocks = vSize / launchCfg.S;
	if(numBlocks * launchCfg.S != vSize)
		return printf("* FIXME: Incomplete block, size%%%i\n", launchCfg.S), -1;

	// Block geometry decomposition (for cards with CUDA caps 2.0)
	int blocksX = numBlocks, blocksY = 1;
	for( ; blocksX > 32768 && blocksY < blocksX; blocksY <<=1, blocksX >>= 1)
		if(blocksX & 0x01) break;         // Can't factorize further by 2^x
	dim3 blocksPerGrid(blocksX, blocksY); // Final block geometry

	// Set cache configuration and launch kernel
	cudaFuncSetCacheConfig((const char*)launchCfg.kerPtr, cudaFuncCachePreferShared);
	// cudaFuncSetSharedMemConfig((const char*)kerPtr, cudaSharedMemBankSizeEightByte);

	
	launchCfg.kerPtr<<<blocksPerGrid, threadsPerBlock>>>((float4*)srcPtr, (float4*) dstPtr, N);
	
	// Cuda kernel calls are asynchronous, sync to check execution errors
	return 0;
}


//---- Interface Functions -----------------------------------------------

//- Main library function for 'float' scan
int KScanLF( float* input, float* output,
	int dir, int N, int stride, int batch)
{
	
	
	if(N>4096)
		return -1;

	return KLauncher(tabla7, sizeof(tabla7),
		input, output, dir, N, batch);
}







